/* Pure CSS "Fork me on GitHub" ribbon.
 * From
 * https://github.com/ssokolow/quicktile/commit/1ae5388ac0f2a2bfa494045644b0ba19eb042329
 * See https://github.com/bitprophet/alabaster/issues/166
 */

#forkongithub {
  position: absolute;
  display: block;
  top: 0;
  right: 0;
  width: 200px;
  overflow: hidden;
  height: 200px;
  z-index: 9999;
}
#forkongithub a {
  background: #000;
  color: #fff;
  text-decoration: none;
  font-family: arial, sans-serif;
  text-align: center;
  font-weight: bold;
  padding: 5px 40px;
  font-size: 10pt;
  line-height: 15pt;
  width: 200px;
  position: absolute;
  top: 40px;
  right: -82px;
  transform: rotate(45deg);
  -webkit-transform: rotate(45deg);
  -ms-transform: rotate(45deg);
  -moz-transform: rotate(45deg);
  -o-transform: rotate(45deg);
  box-shadow: 2px 2px 5px rgba(0, 0, 0, 0.2);
}
#forkongithub a::before,
#forkongithub a::after {
  content: '';
  width: 100%;
  display: block;
  position: absolute;
  top: 1px;
  left: 0;
  height: 1px;
  background: #777;
}
#forkongithub a::after {
  bottom: 1px;
  top: auto;
}
@media screen and (max-width: 979px) {
  #forkongithub a {
    display: none;
  }
}
@media print {
  #forkongithub a {
    display: none;
  }
}
